#include <stdio.h>

#include <util/application.h>
#include <gui/font.h>

#include "qfont.h"

#include <map>
#include <vector>
#include <cassert>

static std::vector<os::Font*> g_cFonts;
static std::vector<uint32> g_anFontFlags;
static os::Locker g_cFontLock( "font_lock" );


class QFontInternal
{
public:
	os::Font* FindFont( os::String cFamily, float vSize, uint32* pnFlags )
	{
		for( uint i = 0; i < g_cFonts.size(); ++i )
		{
			os::Font* pcFont = g_cFonts[i];
			
			if( pcFont->GetFamily() == cFamily && pcFont->GetSize() == vSize )
			{
				pcFont->AddRef();
				//printf( "Fontlist size %i\n", g_cFonts.size() );
				*pnFlags = g_anFontFlags[i];
				return( pcFont );
			}
		}
		return( NULL );
	}
	
	os::Font* CreateDefaultFont( float vSize, uint32* pnFlags )
	{
		os::Font* pcFont = NULL;
		os::font_properties sProp;
		os::Font::GetDefaultFont( DEFAULT_FONT_REGULAR, &sProp );
		
		
		if( ( pcFont = CreateFont( sProp.m_cFamily, vSize, pnFlags, true ) ) )
			return( pcFont );
		
		sProp.m_vSize = vSize;
		pcFont = new os::Font( sProp );
		pcFont->AddRef();
		g_cFonts.push_back( pcFont );
		g_anFontFlags.push_back( *pnFlags );
		printf( "Error: Default Font configuration seems to be invalid!\n");
		return( pcFont );
	}

	os::Font* CreateFont( os::String cFamily, float nSize, uint32* pnFlags, bool bNoDefault = false )
	{
		char zStyleName[FONT_STYLE_LENGTH];
		g_cFontLock.Lock();
		
		os::Font* pcNewFont;
		if( ( pcNewFont = FindFont( cFamily, nSize, pnFlags ) ) )
		{
			g_cFontLock.Unlock();
			return( pcNewFont );
		}
		
		//printf("Searching font %s\n", cFamily.c_str() );
		
		for( int i = 0; i < os::Font::GetStyleCount( cFamily.c_str() ); ++ i )
		{
			uint32 nFlags;
			if( os::Font::GetStyleInfo( cFamily.c_str(), i, zStyleName, &nFlags ) == 0 )
			{
				if( !strcmp( zStyleName, "Regular" ) || !strcmp( zStyleName, "Roman" )
					|| !strcmp( zStyleName, "Book" ) )
				{
					//printf( "Trying Font %s %s %f\n", cFamily.c_str(), zStyleName, nSize );
					
					pcNewFont = new os::Font();
					
					os::font_properties sProp( cFamily, zStyleName, os::FPF_SMOOTHED | os::FPF_SYSTEM, nSize );
					if( pcNewFont->SetProperties( sProp ) == 0 )
					{
						pcNewFont->AddRef();
						//printf( "New font with size %f and flags %x\n", pcNewFont->GetSize(), (uint)nFlags );
						g_cFonts.push_back( pcNewFont );
						g_anFontFlags.push_back( nFlags );
						g_cFontLock.Unlock();
						return( pcNewFont );
					}
					else {
						printf("os::Font::SetProperties() failed!\n");
						pcNewFont->Release();
					}
				}
			}
		}
		if( bNoDefault )
			return( NULL );
		//printf( "Using default font instead of %s\n", cFamily.c_str() );
		pcNewFont = CreateDefaultFont( nSize, pnFlags );
		g_cFontLock.Unlock();
		return( pcNewFont  );
	}
	os::Font*	m_pcFont;
	uint32 m_nFlags;
};


QFont::QFont()
{
    d = new QFontInternal;
    g_cFontLock.Lock();
    d->m_pcFont = d->CreateDefaultFont( 9, &d->m_nFlags );
    g_cFontLock.Unlock();
}
QFont::QFont( const QString& family, int pointSize , int/* weight*/, bool /*italic*/ )
{
    d = new QFontInternal;
    d->m_pcFont = d->CreateFont( family.ascii(), pointSize, &d->m_nFlags );
}

QFont::QFont( const QFont& cOther )
{
    d = new QFontInternal;
    d->m_pcFont = d->CreateFont( cOther.family().ascii(), cOther.pointSize(), &d->m_nFlags );
}

QFont::~QFont()
{
	d->m_pcFont->Release();
    delete d;
}

void QFont::setFamily( const QString &cFont )
{
	if( cFont.ascii() == NULL )
		return;
	float nSize = 9;
	if( d->m_pcFont )
		nSize = d->m_pcFont->GetSize();
	d->m_pcFont->Release();
	d->m_pcFont = d->CreateFont( cFont.ascii(), nSize, &d->m_nFlags );
}

QString QFont::family()	const
{
	assert( d->m_pcFont != NULL );
	return( QString( d->m_pcFont->GetFamily().c_str() ) );
}

void QFont::setCharSet( CharSet )
{
//    printf( "Warning: QFont::%s() not implemented\n", __FUNCTION__ );
}

QFont::CharSet QFont::charSet() const
{
//    printf( "Warning: QFont::%s() not implemented\n", __FUNCTION__ );
    return( ISO_8859_1 );
}
    

bool QFont::italic() const
{
//    printf( "Warning: QFont::%s() not implemented\n", __FUNCTION__ );
    return( false );
}



void QFont::setItalic( bool )
{
    //printf( "Warning: QFont::%s() not implemented\n", __FUNCTION__ );
}

bool QFont::bold() const
{
//    printf( "Warning: QFont::%s() not implemented\n", __FUNCTION__ );
    return( false );
}



void QFont::setBold( bool )
{
    printf( "Warning: QFont::%s() not implemented\n", __FUNCTION__ );
}


int QFont::weight() const
{
//    printf( "Warning: QFont::%s() not implemented\n", __FUNCTION__ );
    return( Normal );
}

void QFont::setWeight( int )
{
//    printf( "Warning: QFont::%s() not implemented\n", __FUNCTION__ );
}

bool QFont::underline() const
{
//    printf( "Warning: QFont::%s() not implemented\n", __FUNCTION__ );
    return( false );
}

bool QFont::overline() const
{
//    printf( "Warning: QFont::%s() not implemented\n", __FUNCTION__ );
    return( false );
}

bool QFont::strikeOut() const
{
//    printf( "Warning: QFont::%s() not implemented\n", __FUNCTION__ );
    return( false );
}


void QFont::setPointSize( int nSize )
{
	os::String cFamily = d->m_pcFont->GetFamily();
	d->m_pcFont->Release();
    d->m_pcFont = d->CreateFont( cFamily, nSize, &d->m_nFlags );
}

void QFont::setPixelSize( int nSize )
{
	 if( nSize == 0 )
	 	return;
	 setPointSize( nSize * 72 / 96 );
}
    
int QFont::pixelSize() const
{
    os::font_height sHeight;
    d->m_pcFont->GetHeight( &sHeight );
    return( int(sHeight.ascender + sHeight.descender/* + sHeight.line_gap + 0.5*/ ) );
}

int QFont::pointSize() const
{
    return( int( d->m_pcFont->GetSize() ) );
}

void QFont::setStyleHint( StyleHint eHint, StyleStrategy )
{
	 printf( "Warning: QFont::%s(%i) not implemented\n", __FUNCTION__, eHint );
}
    
QFont& QFont::operator=( const QFont& cOther )
{
	d->m_pcFont->Release();
    d->m_pcFont = cOther.GetFont();
    d->m_nFlags = cOther.d->m_nFlags;
    assert( d->m_pcFont != NULL );
    d->m_pcFont->AddRef();
    return( *this );
}

bool QFont::operator==( const QFont& cOther ) const
{
    return( *d->m_pcFont == *cOther.d->m_pcFont );
}

bool QFont::operator!=( const QFont& cOther ) const
{
    return( *d->m_pcFont != *cOther.d->m_pcFont );
}


os::Font* QFont::GetFont() const
{
    return( d->m_pcFont );
}


uint32 QFont::GetFontFlags() const
{
    return( d->m_nFlags );
}

void QFont::SetFont( os::Font* pcFont )
{
//    printf( "Set font\n" );
	d->m_pcFont->Release();
    d->m_pcFont = pcFont;
    d->m_nFlags = 0;
    d->m_pcFont->AddRef();
}


























